# Needed for types imported only during TYPE_CHECKING with Python 3.7 - 3.9
# See https://www.python.org/dev/peps/pep-0655/#usage-in-python-3-11
from __future__ import annotations

__all__ = (
    "RenderFunction",
    "RenderFunctionAsync",
    "RenderText",
    "RenderTextAsync",
    "text",
    "RenderPlot",
    "RenderPlotAsync",
    "plot",
    "RenderImage",
    "RenderImageAsync",
    "image",
    "RenderTable",
    "RenderTableAsync",
    "table",
    "RenderUI",
    "RenderUIAsync",
    "ui",
)

import base64
import os
import sys
import typing
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Generic,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)

# These aren't used directly in this file, but they seem necessary for Sphinx to work
# cleanly.
from htmltools import Tag  # pyright: ignore[reportUnusedImport] # noqa: F401
from htmltools import Tagifiable  # pyright: ignore[reportUnusedImport] # noqa: F401
from htmltools import TagList  # pyright: ignore[reportUnusedImport] # noqa: F401
from htmltools import TagChild

if TYPE_CHECKING:
    from ..session import Session
    from ..session._utils import RenderedDeps
    import pandas as pd

from typing import Protocol, runtime_checkable

from .. import _utils
from .._namespaces import ResolvedId
from ..types import ImgData
from ._try_render_plot import try_render_matplotlib, try_render_pil, try_render_plotnine

# Input type for the user-spplied function that is passed to a render.xx
IT = TypeVar("IT")
# Output type after the RenderFunction.__call__ method is called on the IT object.
OT = TypeVar("OT")


# ======================================================================================
# RenderFunction/RenderFunctionAsync base class
# ======================================================================================


# A RenderFunction object is given a user-provided function which returns an IT. When
# the .__call___ method is invoked, it calls the user-provided function (which returns
# an IT), then converts the IT to an OT. Note that in many cases but not all, IT and OT
# will be the same.
class RenderFunction(Generic[IT, OT]):
    def __init__(self, fn: Callable[[], IT]) -> None:
        self.__name__ = fn.__name__
        self.__doc__ = fn.__doc__

    def __call__(self) -> OT:
        raise NotImplementedError

    def set_metadata(self, session: Session, name: str) -> None:
        """When RenderFunctions are assigned to Output object slots, this method
        is used to pass along session and name information.
        """
        self._session: Session = session
        self._name: str = name


# The reason for having a separate RenderFunctionAsync class is because the __call__
# method is marked here as async; you can't have a single class where one method could
# be either sync or async.
class RenderFunctionAsync(RenderFunction[IT, OT]):
    async def __call__(self) -> OT:  # pyright: ignore[reportIncompatibleMethodOverride]
        raise NotImplementedError


# ======================================================================================
# RenderText
# ======================================================================================
RenderTextFunc = Callable[[], "str | None"]
RenderTextFuncAsync = Callable[[], Awaitable["str | None"]]


class RenderText(RenderFunction["str | None", "str | None"]):
    def __init__(self, fn: RenderTextFunc) -> None:
        super().__init__(fn)
        # The Render*Async subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: RenderTextFuncAsync = _utils.wrap_async(fn)

    def __call__(self) -> str | None:
        return _utils.run_coro_sync(self._run())

    async def _run(self) -> str | None:
        res = await self._fn()
        if res is None:
            return None
        return str(res)


class RenderTextAsync(RenderText, RenderFunctionAsync["str | None", "str | None"]):
    def __init__(self, fn: RenderTextFuncAsync) -> None:
        if not _utils.is_async_callable(fn):
            raise TypeError(self.__class__.__name__ + " requires an async function")
        super().__init__(typing.cast(RenderTextFunc, fn))

    async def __call__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
    ) -> str | None:
        return await self._run()


@overload
def text(fn: RenderTextFunc | RenderTextFuncAsync) -> RenderText:
    ...


@overload
def text() -> Callable[[RenderTextFunc | RenderTextFuncAsync], RenderText]:
    ...


def text(
    fn: Optional[RenderTextFunc | RenderTextFuncAsync] = None,
) -> RenderText | Callable[[RenderTextFunc | RenderTextFuncAsync], RenderText]:
    """
    Reactively render text.

    Returns
    -------
    :
        A decorator for a function that returns a string.

    Tip
    ----
    This decorator should be applied **before** the ``@output`` decorator. Also, the
    name of the decorated function (or ``@output(id=...)``) should match the ``id`` of
    a :func:`~shiny.ui.output_text` container (see :func:`~shiny.ui.output_text` for
    example usage).

    See Also
    --------
    ~shiny.ui.output_text
    """

    def wrapper(fn: RenderTextFunc | RenderTextFuncAsync) -> RenderText:
        if _utils.is_async_callable(fn):
            return RenderTextAsync(fn)
        else:
            fn = typing.cast(RenderTextFunc, fn)
            return RenderText(fn)

    if fn is None:
        return wrapper
    else:
        return wrapper(fn)


# ======================================================================================
# RenderPlot
# ======================================================================================
# It would be nice to specify the return type of RenderPlotFunc to be something like:
#   Union[matplotlib.figure.Figure, PIL.Image.Image]
# However, if we did that, we'd have to import those modules at load time, which adds
# a nontrivial amount of overhead. So for now, we're just using `object`.
RenderPlotFunc = Callable[[], object]
RenderPlotFuncAsync = Callable[[], Awaitable[object]]


class RenderPlot(RenderFunction[object, "ImgData | None"]):
    _ppi: float = 96
    _is_userfn_async = False

    def __init__(
        self, fn: RenderPlotFunc, *, alt: Optional[str] = None, **kwargs: object
    ) -> None:
        super().__init__(fn)
        self._alt: Optional[str] = alt
        self._kwargs = kwargs
        # The Render*Async subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: RenderPlotFuncAsync = _utils.wrap_async(fn)

    def __call__(self) -> ImgData | None:
        return _utils.run_coro_sync(self._run())

    async def _run(self) -> ImgData | None:
        inputs = self._session.root_scope().input

        # Reactively read some information about the plot.
        pixelratio: float = typing.cast(
            float, inputs[ResolvedId(".clientdata_pixelratio")]()
        )
        width: float = typing.cast(
            float, inputs[ResolvedId(f".clientdata_output_{self._name}_width")]()
        )
        height: float = typing.cast(
            float, inputs[ResolvedId(f".clientdata_output_{self._name}_height")]()
        )

        x = await self._fn()

        # Note that x might be None; it could be a matplotlib.pyplot

        # Try each type of renderer in turn. The reason we do it this way is to avoid
        # importing modules that aren't already loaded. That could slow things down, or
        # worse, cause an error if the module isn't installed.
        #
        # Each try_render function should indicate whether it was able to make sense of
        # the x value (or, in the case of matplotlib, possibly it decided to use the
        # global pyplot figure) by returning a tuple that starts with True. The second
        # tuple element may be None in this case, which means the try_render function
        # explicitly wants the plot to be blanked.
        #
        # If a try_render function returns a tuple that starts with False, then the next
        # try_render function should be tried. If none succeed, an error is raised.
        ok: bool
        result: ImgData | None

        if "plotnine" in sys.modules:
            ok, result = try_render_plotnine(
                x, width, height, pixelratio, self._ppi, **self._kwargs
            )
            if ok:
                return result

        if "matplotlib" in sys.modules:
            ok, result = try_render_matplotlib(
                x,
                width,
                height,
                pixelratio=pixelratio,
                ppi=self._ppi,
                allow_global=not self._is_userfn_async,
                alt=self._alt,
                **self._kwargs,
            )
            if ok:
                return result

        if "PIL" in sys.modules:
            ok, result = try_render_pil(
                x, width, height, pixelratio, self._ppi, **self._kwargs
            )
            if ok:
                return result

        # This check must happen last because matplotlib might be able to plot even if
        # x is None
        if x is None:
            return None

        raise Exception(
            f"@render.plot doesn't know to render objects of type '{str(type(x))}'. "
            + "Consider either requesting support for this type of plot object, and/or "
            + " explictly saving the object to a (png) file and using @render.image."
        )


class RenderPlotAsync(RenderPlot, RenderFunctionAsync[object, "ImgData | None"]):
    _is_userfn_async = True

    def __init__(
        self, fn: RenderPlotFuncAsync, alt: Optional[str] = None, **kwargs: Any
    ) -> None:
        if not _utils.is_async_callable(fn):
            raise TypeError(self.__class__.__name__ + " requires an async function")
        super().__init__(typing.cast(RenderPlotFunc, fn), alt=alt, **kwargs)

    async def __call__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
    ) -> ImgData | None:
        return await self._run()


@overload
def plot(fn: RenderPlotFunc | RenderPlotFuncAsync) -> RenderPlot:
    ...


@overload
def plot(
    *,
    alt: Optional[str] = None,
    **kwargs: Any,
) -> Callable[[RenderPlotFunc | RenderPlotFuncAsync], RenderPlot]:
    ...


# TODO: Use more specific types for render.plot
def plot(
    fn: Optional[RenderPlotFunc | RenderPlotFuncAsync] = None,
    *,
    alt: Optional[str] = None,
    **kwargs: Any,
) -> RenderPlot | Callable[[RenderPlotFunc | RenderPlotFuncAsync], RenderPlot]:
    """
    Reactively render a plot object as an HTML image.

    Parameters
    ----------
    alt
        Alternative text for the image if it cannot be displayed or viewed (i.e., the
        user uses a screen reader).
    **kwargs
        Additional keyword arguments passed to the relevant method for saving the image
        (e.g., for matplotlib, arguments to ``savefig()``; for PIL and plotnine,
        arguments to ``save()``).

    Returns
    -------
    :
        A decorator for a function that returns any of the following:

        1. A :class:`matplotlib.figure.Figure` instance.
        2. An :class:`matplotlib.artist.Artist` instance.
        3. A list/tuple of Figure/Artist instances.
        4. An object with a 'figure' attribute pointing to a
           :class:`matplotlib.figure.Figure` instance.
        5. A :class:`PIL.Image.Image` instance.

    It's also possible to use the ``matplotlib.pyplot`` interface; in that case, your
    function should just call pyplot functions and not return anything. (Note that if
    the decorated function is async, then it's not safe to use pyplot. Shiny will detect
    this case and throw an error asking you to use matplotlib's object-oriented
    interface instead.)

    Tip
    ----
    This decorator should be applied **before** the ``@output`` decorator. Also, the
    name of the decorated function (or ``@output(id=...)``) should match the ``id`` of a
    :func:`~shiny.ui.output_plot` container (see :func:`~shiny.ui.output_plot` for
    example usage).

    See Also
    --------
    ~shiny.ui.output_plot
    ~shiny.render.image
    """

    def wrapper(fn: RenderPlotFunc | RenderPlotFuncAsync) -> RenderPlot:
        if _utils.is_async_callable(fn):
            return RenderPlotAsync(fn, alt=alt, **kwargs)
        else:
            return RenderPlot(fn, alt=alt, **kwargs)

    if fn is None:
        return wrapper
    else:
        return wrapper(fn)


# ======================================================================================
# RenderImage
# ======================================================================================
RenderImageFunc = Callable[[], "ImgData | None"]
RenderImageFuncAsync = Callable[[], Awaitable["ImgData | None"]]


class RenderImage(RenderFunction["ImgData | None", "ImgData | None"]):
    def __init__(
        self,
        fn: RenderImageFunc,
        *,
        delete_file: bool = False,
    ) -> None:
        super().__init__(fn)
        self._delete_file: bool = delete_file
        # The Render*Async subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: RenderImageFuncAsync = _utils.wrap_async(fn)

    def __call__(self) -> ImgData | None:
        return _utils.run_coro_sync(self._run())

    async def _run(self) -> ImgData | None:
        res: ImgData | None = await self._fn()
        if res is None:
            return None

        src: str = res.get("src")
        try:
            with open(src, "rb") as f:
                data = base64.b64encode(f.read())
                data_str = data.decode("utf-8")
            content_type = _utils.guess_mime_type(src)
            res["src"] = f"data:{content_type};base64,{data_str}"
            return res
        finally:
            if self._delete_file:
                os.remove(src)


class RenderImageAsync(
    RenderImage, RenderFunctionAsync["ImgData | None", "ImgData | None"]
):
    def __init__(self, fn: RenderImageFuncAsync, delete_file: bool = False) -> None:
        if not _utils.is_async_callable(fn):
            raise TypeError(self.__class__.__name__ + " requires an async function")
        super().__init__(typing.cast(RenderImageFunc, fn), delete_file=delete_file)

    async def __call__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
    ) -> ImgData | None:
        return await self._run()


@overload
def image(fn: RenderImageFunc | RenderImageFuncAsync) -> RenderImage:
    ...


@overload
def image(
    *,
    delete_file: bool = False,
) -> Callable[[RenderImageFunc | RenderImageFuncAsync], RenderImage]:
    ...


def image(
    fn: Optional[RenderImageFunc | RenderImageFuncAsync] = None,
    *,
    delete_file: bool = False,
) -> RenderImage | Callable[[RenderImageFunc | RenderImageFuncAsync], RenderImage]:
    """
    Reactively render a image file as an HTML image.

    Parameters
    ----------
    delete_file
        If ``True``, the image file will be deleted after rendering.

    Returns
    -------
    :
        A decorator for a function that returns an `~shiny.types.ImgData` object.

    Tip
    ----
    This decorator should be applied **before** the ``@output`` decorator. Also, the
    name of the decorated function (or ``@output(id=...)``) should match the ``id`` of
    a :func:`~shiny.ui.output_image` container (see :func:`~shiny.ui.output_image` for
    example usage).

    See Also
    --------
    ~shiny.ui.output_image
    ~shiny.types.ImgData
    ~shiny.render.plot
    """

    def wrapper(fn: RenderImageFunc | RenderImageFuncAsync) -> RenderImage:
        if _utils.is_async_callable(fn):
            return RenderImageAsync(fn, delete_file=delete_file)
        else:
            fn = typing.cast(RenderImageFunc, fn)
            return RenderImage(fn, delete_file=delete_file)

    if fn is None:
        return wrapper
    else:
        return wrapper(fn)


# ======================================================================================
# RenderTable
# ======================================================================================


@runtime_checkable
class PandasCompatible(Protocol):
    # Signature doesn't matter, runtime_checkable won't look at it anyway
    def to_pandas(self) -> "pd.DataFrame":
        ...


TableResult = Union[None, "pd.DataFrame", PandasCompatible]
RenderTableFunc = Callable[[], TableResult]
RenderTableFuncAsync = Callable[[], Awaitable[TableResult]]


class RenderTable(RenderFunction[object, "RenderedDeps | None"]):
    def __init__(
        self,
        fn: RenderTableFunc,
        *,
        index: bool = False,
        classes: str = "table shiny-table w-auto",
        border: int = 0,
        **kwargs: object,
    ) -> None:
        super().__init__(fn)
        self._index = index
        self._classes = classes
        self._border = border
        self._kwargs = kwargs
        # The Render*Async subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: RenderTableFuncAsync = _utils.wrap_async(fn)

    def __call__(self) -> RenderedDeps | None:
        return _utils.run_coro_sync(self._run())

    async def _run(self) -> RenderedDeps | None:
        x = await self._fn()

        if x is None:
            return None

        import pandas
        import pandas.io.formats.style

        html: str
        if isinstance(x, pandas.io.formats.style.Styler):
            html = x.to_html(**self._kwargs)  # pyright: ignore[reportUnknownMemberType]
        else:
            if not isinstance(x, pandas.DataFrame):
                if not isinstance(x, PandasCompatible):
                    raise TypeError(
                        "@render.table doesn't know how to render objects of type "
                        f"'{str(type(x))}'. Return either a pandas.DataFrame, or an object "
                        "that has a .to_pandas() method."
                    )
                x = x.to_pandas()

            html = x.to_html(  # pyright: ignore[reportUnknownMemberType]
                index=self._index,
                classes=self._classes,
                border=self._border,
                **self._kwargs,
            )
        return {"deps": [], "html": html}


class RenderTableAsync(RenderTable, RenderFunctionAsync[object, "ImgData | None"]):
    def __init__(
        self,
        fn: RenderTableFuncAsync,
        *,
        index: bool = False,
        classes: str = "table shiny-table w-auto",
        border: int = 0,
        **kwargs: Any,
    ) -> None:
        if not _utils.is_async_callable(fn):
            raise TypeError(self.__class__.__name__ + " requires an async function")
        super().__init__(
            typing.cast(RenderTableFunc, fn),
            index=index,
            classes=classes,
            border=border,
            **kwargs,
        )

    async def __call__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
    ) -> RenderedDeps | None:
        return await self._run()


@overload
def table(fn: RenderTableFunc | RenderTableFuncAsync) -> RenderTable:
    ...


@overload
def table(
    *,
    index: bool = False,
    classes: str = "table shiny-table w-auto",
    border: int = 0,
    **kwargs: Any,
) -> Callable[[RenderTableFunc | RenderTableFuncAsync], RenderTable]:
    ...


# TODO: Use more specific types for render.table
def table(
    fn: Optional[RenderTableFunc | RenderTableFuncAsync] = None,
    *,
    index: bool = False,
    classes: str = "table shiny-table w-auto",
    border: int = 0,
    **kwargs: Any,
) -> RenderTable | Callable[[RenderTableFunc | RenderTableFuncAsync], RenderTable]:
    """
    Reactively render a Pandas data frame object (or similar) as a basic HTML table.

    Consider using ~shiny.render.data_frame instead of this renderer, as it provides
    high performance virtual scrolling, built-in filtering and sorting, and a better
    default appearance. This renderer may still be helpful if you use pandas styling
    features that are not currently supported by ~shiny.render.data_frame.

    Parameters
    ----------
    index
        Whether to print index (row) labels. (Ignored for pandas :class:`Styler`
        objects; call ``style.hide(axis="index")`` from user code instead.)
    classes
        CSS classes (space separated) to apply to the resulting table. By default, we
        use `table shiny-table w-auto` which is designed to look reasonable with Bootstrap 5.
        (Ignored for pandas :class:`Styler` objects; call
        ``style.set_table_attributes('class="dataframe table shiny-table w-auto"')``
        from user code instead.)
    **kwargs
        Additional keyword arguments passed to ``pandas.DataFrame.to_html()`` or
        ``pandas.io.formats.style.Styler.to_html()``.

    Returns
    -------
    :
        A decorator for a function that returns any of the following:

        1. A pandas :class:`DataFrame` object.
        2. A pandas :class:`Styler` object.
        3. Any object that has a `.to_pandas()` method (e.g., a Polars data frame or
           Arrow table).

    Tip
    ----
    This decorator should be applied **before** the ``@output`` decorator. Also, the
    name of the decorated function (or ``@output(id=...)``) should match the ``id`` of
    a :func:`~shiny.ui.output_table` container (see :func:`~shiny.ui.output_table` for
    example usage).

    See Also
    --------
    ~shiny.ui.output_table for the corresponding UI component to this render function.
    """

    def wrapper(fn: RenderTableFunc | RenderTableFuncAsync) -> RenderTable:
        if _utils.is_async_callable(fn):
            return RenderTableAsync(
                fn, index=index, classes=classes, border=border, **kwargs
            )
        else:
            return RenderTable(
                cast(RenderTableFunc, fn),
                index=index,
                classes=classes,
                border=border,
                **kwargs,
            )

    if fn is None:
        return wrapper
    else:
        return wrapper(fn)


# ======================================================================================
# RenderUI
# ======================================================================================
RenderUIFunc = Callable[[], TagChild]
RenderUIFuncAsync = Callable[[], Awaitable[TagChild]]


class RenderUI(RenderFunction[TagChild, "RenderedDeps | None"]):
    def __init__(self, fn: RenderUIFunc) -> None:
        super().__init__(fn)
        # The Render*Async subclass will pass in an async function, but it tells the
        # static type checker that it's synchronous. wrap_async() is smart -- if is
        # passed an async function, it will not change it.
        self._fn: RenderUIFuncAsync = _utils.wrap_async(fn)

    def __call__(self) -> RenderedDeps | None:
        return _utils.run_coro_sync(self._run())

    async def _run(self) -> RenderedDeps | None:
        ui: TagChild = await self._fn()
        if ui is None:
            return None

        return self._session._process_ui(ui)


class RenderUIAsync(RenderUI, RenderFunctionAsync[TagChild, "RenderedDeps| None"]):
    def __init__(self, fn: RenderUIFuncAsync) -> None:
        if not _utils.is_async_callable(fn):
            raise TypeError(self.__class__.__name__ + " requires an async function")
        super().__init__(typing.cast(RenderUIFunc, fn))

    async def __call__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
    ) -> RenderedDeps | None:
        return await self._run()


@overload
def ui(fn: RenderUIFunc | RenderUIFuncAsync) -> RenderUI:
    ...


@overload
def ui() -> Callable[[RenderUIFunc | RenderUIFuncAsync], RenderUI]:
    ...


def ui(
    fn: Optional[RenderUIFunc | RenderUIFuncAsync] = None,
) -> RenderUI | Callable[[RenderUIFunc | RenderUIFuncAsync], RenderUI]:
    """
    Reactively render HTML content.

    Returns
    -------
    :
        A decorator for a function that returns an object of type `~shiny.ui.TagChild`.

    Tip
    ----
    This decorator should be applied **before** the ``@output`` decorator. Also, the
    name of the decorated function (or ``@output(id=...)``) should match the ``id`` of
    a :func:`~shiny.ui.output_ui` container (see :func:`~shiny.ui.output_ui` for example
    usage).

    See Also
    --------
    ~shiny.ui.output_ui
    """

    def wrapper(
        fn: Callable[[], TagChild] | Callable[[], Awaitable[TagChild]]
    ) -> RenderUI:
        if _utils.is_async_callable(fn):
            return RenderUIAsync(fn)
        else:
            fn = typing.cast(RenderUIFunc, fn)
            return RenderUI(fn)

    if fn is None:
        return wrapper
    else:
        return wrapper(fn)



import shiny.experimental as x


def test_panel_main_and_panel_sidebar():
    ps = x.ui.layout_sidebar(
        x.ui.panel_sidebar("Sidebar!"),
        x.ui.panel_main("Main!", TestAttr=True),
    )
    ps_str = str(ps.tagify())
    assert "TestAttr" in ps_str
    assert "Sidebar!" in ps_str
    assert "Main!" in ps_str

    try:
        x.ui.layout_sidebar(
            None,
            x.ui.sidebar("Sidebar!"),  # type: ignore
        )
        raise AssertionError("Should have raised TypeError")
    except TypeError as e:
        assert "Please use the `sidebar=` argument" in str(e)

    try:
        x.ui.layout_sidebar(
            None,  # sidebar
            x.ui.panel_sidebar("Sidebar!"),
            x.ui.panel_sidebar("Sidebar!"),
        )
        raise AssertionError("Should have raised TypeError")
    except TypeError as e:
        assert "Multiple `panel_sidebar()` calls detected" in str(e)

    try:
        x.ui.layout_sidebar(
            "Sidebar2!",
            x.ui.panel_sidebar("Sidebar!"),
        )
        raise AssertionError("Should have raised TypeError")
    except TypeError as e:
        assert "was supplied along with" in str(e)



from __future__ import annotations

import shiny.experimental as x
from shiny import App, Inputs, Outputs, Session, reactive, ui

items = [
    x.ui.accordion_panel(f"Section {letter}", f"Some narrative for section {letter}")
    for letter in "ABCDE"
]

app_ui = ui.page_fluid(
    ui.input_action_button("open_acc", "Open Section C", class_="mt-3 mb-3"),
    x.ui.accordion(*items, id="acc", multiple=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.Effect
    @reactive.event(input.open_acc)
    def _():
        x.ui.accordion_panel_open("acc", "Section C")


app = App(app_ui, server)



from __future__ import annotations

import random
from typing import Literal, Optional

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild, TagList, css, div
from htmltools import svg as svgtags
from htmltools import tags

# from ._color import get_color_contrast
from ._css_unit import CssUnit, as_css_padding, as_css_unit
from ._fill import as_fill_item, as_fillable_container
from ._htmldeps import sidebar_dependency
from ._utils import consolidate_attrs, trinary


class Sidebar:
    def __init__(
        self,
        tag: Tag,
        collapse_tag: Optional[Tag],
        position: Literal["left", "right"],
        open: Literal["desktop", "open", "closed", "always"],
        width: CssUnit,
        max_height_mobile: Optional[str | float],
        color_fg: Optional[str],
        color_bg: Optional[str],
    ):
        self.tag = tag
        self.collapse_tag = collapse_tag
        self.position = position
        self.open = open
        self.width = width
        self.max_height_mobile = max_height_mobile
        self.color_fg = color_fg
        self.color_bg = color_bg

    # # This does not contain the `collapse_tag`
    # # The `Sidebar` class should use it's fields, not this method
    # def tagify(self) -> Tag:
    #     return self.tag.tagify()


def sidebar(
    *args: TagChild | TagAttrs,
    width: CssUnit = 250,
    position: Literal["left", "right"] = "left",
    open: Literal["desktop", "open", "closed", "always"] = "desktop",
    id: Optional[str] = None,
    title: TagChild | str = None,
    bg: Optional[str] = None,
    fg: Optional[str] = None,
    class_: Optional[str] = None,  # TODO-future; Consider using `**kwargs` instead
    max_height_mobile: Optional[str | float] = None,
) -> Sidebar:
    # See [this article](https://rstudio.github.io/bslib/articles/sidebars.html)
    #   to learn more.
    # TODO-future; If color contrast is implemented. Docs for `bg` and `fg`:
    #     If only one of either is provided, an
    #     accessible contrasting color is provided for the opposite color, e.g. setting
    #     `bg` chooses an appropriate `fg` color.
    # TODO-future; validate `open`, bg, fg, class_, max_height_mobile

    if id is None and open != "always":
        # but always provide id when collapsible for accessibility reasons
        id = f"bslib-sidebar-{random.randint(1000, 10000)}"

    # TODO-future; implement
    # if fg is None and bg is not None:
    #     fg = get_color_contrast(bg)
    # if bg is None and fg is not None:
    #     bg = get_color_contrast(fg)

    if isinstance(title, (str, int, float)):
        title = div(str(title), class_="sidebar-title")

    collapse_tag = None
    # Code
    if open != "always":
        collapse_tag = tags.button(
            _collapse_icon(),
            class_="collapse-toggle",
            type="button",
            title="Toggle sidebar",
            aria_expanded=trinary(open in ["open", "desktop"]),
            aria_controls=id,
        )

    tag = div(
        div(
            title,
            {"class": "sidebar-content"},
            *args,
        ),
        {"class": "bslib-sidebar-input"} if id is not None else None,
        {"class": "sidebar"},
        id=id,
        role="complementary",
        class_=class_,
    )

    return Sidebar(
        tag=tag,
        collapse_tag=collapse_tag,
        position=position,
        open=open,
        width=width,
        max_height_mobile=max_height_mobile,
        color_fg=fg,
        color_bg=bg,
    )


# TODO-maindocs; @add_example()
def layout_sidebar(
    sidebar: Sidebar,
    content: PanelMain,
    fillable: bool = True,
    fill: bool = True,
    bg: Optional[str] = None,
    fg: Optional[str] = None,
    border: Optional[bool] = None,
    border_radius: Optional[bool] = None,
    border_color: Optional[str] = None,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    height: Optional[CssUnit] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    attrs, _ = consolidate_attrs(**content.attrs, **kwargs)

    main = div(
        {
            "role": "main",
            "class": f"main{' bslib-gap-spacing' if fillable else ''}",
            ""
            "style": css(
                background_color=bg,
                color=fg,
                gap=as_css_unit(gap),
                padding=as_css_padding(padding),
            ),
        },
        attrs,
        content,
    )
    if fillable:
        main = as_fillable_container(main)

    max_height_mobile = sidebar.max_height_mobile or (
        "250px" if height is None else "50%"
    )

    res = div(
        {"class": "bslib-sidebar-layout"},
        {"class": "sidebar-right"} if sidebar.position == "right" else None,
        {"class": "sidebar-collapsed"} if sidebar.open == "closed" else None,
        main,
        sidebar.tag,
        sidebar.collapse_tag,
        sidebar_dependency(),
        _sidebar_init_js(),
        data_bslib_sidebar_init="true" if sidebar.open != "always" else None,
        data_bslib_sidebar_open=sidebar.open,
        data_bslib_sidebar_border=trinary(border),
        data_bslib_sidebar_border_radius=trinary(border_radius),
        style=css(
            __bslib_sidebar_width=as_css_unit(sidebar.width),
            __bslib_sidebar_bg=as_css_unit(sidebar.color_bg),
            __bslib_sidebar_fg=as_css_unit(sidebar.color_fg),
            __bs_card_border_color=border_color,
            height=as_css_unit(height),
            __bslib_sidebar_max_height_mobile=as_css_unit(max_height_mobile),
        ),
    )
    if fill:
        res = as_fill_item(res)

    return res


# _sidebar_func = sidebar


def _collapse_icon() -> Tag:
    return tags.svg(
        svgtags.path(
            fill_rule="evenodd",
            d="M1.646 4.646a.5.5 0 0 1 .708 0L8 10.293l5.646-5.647a.5.5 0 0 1 .708.708l-6 6a.5.5 0 0 1-.708 0l-6-6a.5.5 0 0 1 0-.708z",
        ),
        xmlns="http://www.w3.org/2000/svg",
        viewBox="0 0 16 16",
        class_="bi bi-chevron-down collapse-icon",
        style="fill:currentColor;",
        aria_hidden="true",
        role="img",
    )


def _sidebar_init_js() -> Tag:
    # Note: if we want to avoid inline `<script>` tags in the future for
    # initialization code, we might be able to do so by turning the sidebar layout
    # container into a web component
    return tags.script(
        {"data-bslib-sidebar-init": True},
        "bslib.Sidebar.initCollapsibleAll()",
    )


###################################################################


class PanelSidebar:
    # Store `attrs` for `layout_sidebar()` to retrieve
    def __init__(
        self, *args: TagChild | TagAttrs, width: int = 4, **kwargs: TagAttrValue
    ) -> None:
        self.args = args
        self.kwargs = kwargs
        self.width = width

    def get_sidebar(self, position: Literal["left", "right"] = "left") -> Sidebar:
        return sidebar(
            *self.args,
            width=f"{int(self.width / 12 * 100)}%",
            position=position,
            open="always",
            **self.kwargs,
        )

    # Hopefully this is never used. But it makes it Tagifiable to allow us to not expose
    # `Sidebar` and `PanelSidebar` classes
    def tagify(self) -> Tag:
        return self.get_sidebar().tag.tagify()


class PanelMain:
    # Store `attrs` for `layout_sidebar()` to retrieve
    attrs: TagAttrs
    # Return `children` in `layout_sidebar()` via `.tagify()` method
    children: list[TagChild]

    def __init__(self, *, attrs: TagAttrs, children: list[TagChild]) -> None:
        self.attrs = attrs
        self.children = children

    def tagify(self) -> TagList:
        return TagList(self.children).tagify()



from __future__ import annotations

from pathlib import PurePath
from typing import Literal

from htmltools import HTMLDependency

from ... import __version__ as shiny_version
from ..._typing_extensions import NotRequired, TypedDict
from ..._versions import bslib as bslib_version
from ..._versions import htmltools as htmltools_version

_x_www = PurePath(__file__).parent.parent / "www"
_x_www_path = str(_x_www)
_x_htmltools_path = _x_www / "htmltools"
_x_components_path = _x_www / "bslib" / "components"


def _htmltools_dep(
    name: str,
    script: bool = False,
    stylesheet: bool = False,
    all_files: bool = True,
) -> HTMLDependency:
    return HTMLDependency(
        name=f"htmltools-{name}",
        version=htmltools_version,
        source={
            "package": "shiny",
            "subdir": str(_x_htmltools_path / "fill"),
        },
        script={"src": f"{name}.min.js"} if script else None,
        stylesheet={"href": f"{name}.css"} if stylesheet else None,
        all_files=all_files,
    )


class _ScriptItemDict(TypedDict):
    src: str
    type: NotRequired[Literal["module"]]


def _bslib_component_dep(
    name: str,
    script: bool = False,
    stylesheet: bool = False,
    all_files: bool = True,
    script_is_module: bool = False,
) -> HTMLDependency:
    script_val: _ScriptItemDict | None = None
    if script:
        script_val = {"src": f"{name}.min.js"}
        if script_is_module:
            script_val["type"] = "module"

    return HTMLDependency(
        name=f"bslib-{name}",
        version=bslib_version,
        source={
            "package": "shiny",
            "subdir": str(_x_components_path / name),
        },
        script=script_val,  # type: ignore # https://github.com/rstudio/py-htmltools/issues/59
        stylesheet={"href": f"{name}.css"} if stylesheet else None,
        all_files=all_files,
    )


# -- htmltools ---------------------


def fill_dependency() -> HTMLDependency:
    return _htmltools_dep("fill", stylesheet=True)


# -- bslib -------------------------


def accordion_dependency() -> HTMLDependency:
    return _bslib_component_dep("accordion", script=True, stylesheet=True)


def card_dependency() -> HTMLDependency:
    return _bslib_component_dep("card", script=True, stylesheet=True)


def grid_dependency() -> HTMLDependency:
    return _bslib_component_dep("grid", stylesheet=True)


# Coped to `ui._x._htmldeps.py`
# def nav_spacer_dependency() -> HTMLDependency:
#     return _bslib_component_dep("nav_spacer", stylesheet=True)


def page_fillable_dependency() -> HTMLDependency:
    return _bslib_component_dep("page_fillable", stylesheet=True)


# # Not used!
# def page_navbar_dependency() -> HTMLDependency:
#     return _bslib_component_dep("page_navbar", stylesheet=True)


def page_sidebar_dependency() -> HTMLDependency:
    return _bslib_component_dep("page_sidebar", stylesheet=True)


def sidebar_dependency() -> HTMLDependency:
    return _bslib_component_dep("sidebar", script=True, stylesheet=True)


def value_box_dependency() -> HTMLDependency:
    return _bslib_component_dep("value_box", stylesheet=True)


def web_component_dependency() -> HTMLDependency:
    return _bslib_component_dep("webComponents", script=True, script_is_module=True)


# -- Experimental ------------------


def autoresize_dependency():
    return HTMLDependency(
        "shiny-textarea-autoresize",
        shiny_version,
        source={"package": "shiny", "subdir": _x_www_path},
        script={"src": "textarea-autoresize.js"},
        stylesheet={"href": "textarea-autoresize.css"},
    )
